import re
from datetime import date

import pytest

from django_afip import factories
from django_afip import models
from django_afip.pdf import ReceiptQrCode


@pytest.mark.django_db
def test_pdf_generation():
    """Test PDF file generation.

    For the moment, this test case mostly verifies that pdf generation
    *works*, but does not actually validate the pdf file itself.

    Running this locally *will* yield the file itself, which is useful for
    manual inspection.
    """
    pdf = factories.ReceiptPDFFactory(receipt__receipt_number=3)
    factories.ReceiptValidationFactory(receipt=pdf.receipt)
    pdf.save_pdf()
    regex = r"afip/receipts/[a-f0-9]{2}/[a-f0-9]{2}/[a-f0-9]{32}.pdf"

    assert re.match(regex, pdf.pdf_file.name)
    assert pdf.pdf_file.name.endswith(".pdf")


@pytest.mark.django_db
def test_unauthorized_receipt_generation():
    """
    Test PDF file generation for unauthorized receipts.

    Confirm that attempting to generate a PDF for an unauthorized receipt
    raises.
    """
    taxpayer = factories.TaxPayerFactory()
    receipt = factories.ReceiptFactory(
        receipt_number=None,
        point_of_sales__owner=taxpayer,
    )
    pdf = models.ReceiptPDF.objects.create_for_receipt(
        receipt=receipt,
        client_name="John Doe",
        client_address="12 Green Road\nGreenville\nUK",
    )
    with pytest.raises(
        Exception, match="Cannot generate pdf for non-authorized receipt"
    ):
        pdf.save_pdf()


@pytest.mark.django_db
def test_signal_generation_for_not_validated_receipt():
    printable = factories.ReceiptPDFFactory()

    assert not (printable.pdf_file)


@pytest.mark.django_db
def test_signal_generation_for_validated_receipt():
    validation = factories.ReceiptValidationFactory()
    printable = factories.ReceiptPDFFactory(receipt=validation.receipt)

    assert printable.pdf_file
    assert printable.pdf_file.name.endswith(".pdf")


@pytest.mark.django_db
def test_qrcode_data():
    pdf = factories.ReceiptPDFFactory(
        receipt__receipt_number=3,
        receipt__issued_date=date(2021, 3, 2),
    )
    factories.ReceiptValidationFactory(receipt=pdf.receipt)

    qrcode = ReceiptQrCode(pdf.receipt)
    assert qrcode._data == {
        "codAut": "67190616790549",
        "ctz": 1.0,
        "cuit": "20329642330",
        "fecha": "2021-03-02",
        "importe": 130.0,
        "moneda": "PES",
        "nroCmp": 3,
        "nroDocRec": 203012345,
        "ptoVta": 1,
        "tipoCmp": 6,
        "tipoCodAut": "E",
        "tipoDocRec": 96,
        "ver": 1,
    }



import re

from django import template

register = template.Library()


@register.filter
def format_cuit(cuit):
    numbers = re.sub("[^\\d]", "", str(cuit))
    if len(numbers) != 11:
        return cuit
    return f"{numbers[0:2]}-{numbers[2:10]}-{numbers[10:11]}"



import django.db.models.deletion
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ("afip", "0001_squashed_0036_receiptpdf__client_address__blank"),
    ]

    operations = [
        migrations.CreateModel(
            name="TaxPayerExtras",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "logo",
                    models.ImageField(
                        blank=True,
                        help_text="A logo to use when generating printable receipts.",
                        null=True,
                        upload_to="afip/taxpayers/logos/",
                        verbose_name="pdf file",
                    ),
                ),
                (
                    "taxpayer",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="extras",
                        to="afip.TaxPayer",
                        verbose_name="taxpayer",
                    ),
                ),
            ],
            options={
                "verbose_name": "taxpayer extras",
                "verbose_name_plural": "taxpayers extras",
            },
        ),
    ]



from django.core.management.base import BaseCommand
from django.utils.translation import gettext as _

from django_afip import models


class Command(BaseCommand):
    help = _("Loads fixtures with metadata from AFIP.")
    requires_migrations_checks = True

    def handle(self, *args, **options):
        models.load_metadata()



import os
from pathlib import Path

import dj_database_url

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = Path(__file__).resolve().parent.parent


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.8/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "dm=3#ff_%6wpr8ynq!r_+kdl$c!$gez$@x!o==k1k96im4ckk+"

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS: list = []


# Application definition

INSTALLED_APPS = (
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "testapp.testmain",
    "django_afip",
)

MIDDLEWARE = (
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "django.middleware.security.SecurityMiddleware",
)

ROOT_URLCONF = "testapp.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

WSGI_APPLICATION = "testapp.wsgi.application"

# Database
DATABASES = {"default": dj_database_url.config()}

# Internationalization
# https://docs.djangoproject.com/en/1.8/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.8/howto/static-files/

STATIC_URL = "/static/"
STATIC_ROOT = "/tmp/static/"

MEDIA_URL = "/media/"
MEDIA_ROOT = os.path.join(BASE_DIR, "media/")



from django.db.models.signals import post_save
from django.db.models.signals import pre_save
from django.dispatch import receiver

from django_afip import models


@receiver(pre_save, sender=models.TaxPayer)
def update_certificate_expiration(sender, instance: models.TaxPayer, **kwargs):
    if instance.certificate:
        instance.certificate_expiration = instance.get_certificate_expiration()


@receiver(post_save, sender=models.ReceiptPDF)
def generate_receipt_pdf(sender, instance: models.ReceiptPDF, **kwargs):
    if not instance.pdf_file and instance.receipt.is_validated:
        instance.save_pdf(save_model=True)



from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.serialization import Encoding
from cryptography.hazmat.primitives.serialization import load_pem_private_key
from cryptography.hazmat.primitives.serialization.pkcs7 import PKCS7Options
from cryptography.hazmat.primitives.serialization.pkcs7 import PKCS7SignatureBuilder
from cryptography.x509 import load_pem_x509_certificate
from OpenSSL import crypto

from django_afip import exceptions


def create_embeded_pkcs7_signature(data: bytes, cert: bytes, key: bytes) -> bytes:
    """Creates an embedded ("nodetached") PKCS7 signature.

    This is equivalent to the output of::

        openssl smime -sign -signer cert -inkey key -outform DER -nodetach < data
    """

    try:
        pkey = load_pem_private_key(key, None)
        signcert = load_pem_x509_certificate(cert)
    except Exception as e:
        raise exceptions.CorruptCertificate from e

    signed_data = (
        PKCS7SignatureBuilder()
        .set_data(data)
        .add_signer(signcert, pkey, hashes.SHA256())
        .sign(Encoding.DER, [PKCS7Options.Binary])
    )
    return signed_data


def create_key(file_):
    """
    Create a key and save it into ``file_``.

    Note that ``file`` must be opened in binary mode.
    """
    pkey = crypto.PKey()
    pkey.generate_key(crypto.TYPE_RSA, 2048)

    file_.write(crypto.dump_privatekey(crypto.FILETYPE_PEM, pkey))
    file_.flush()


def create_csr(key_file, organization_name, common_name, serial_number, file_):
    """Create a CSR for a key, and save it into ``file_``."""
    key = crypto.load_privatekey(crypto.FILETYPE_PEM, key_file.read())

    req = crypto.X509Req()
    subj = req.get_subject()

    subj.O = organization_name  # noqa: E741 (we can't do anything about this)
    subj.CN = common_name
    subj.serialNumber = serial_number

    req.set_pubkey(key)
    req.sign(key, "md5")

    file_.write(crypto.dump_certificate_request(crypto.FILETYPE_PEM, req))



from datetime import date
from datetime import datetime
from pathlib import Path

from django.contrib.auth.models import User
from django.utils.timezone import make_aware
from factory import LazyFunction
from factory import PostGenerationMethodCall
from factory import Sequence
from factory import SubFactory
from factory import post_generation
from factory.django import DjangoModelFactory
from factory.django import FileField
from factory.django import ImageField

from django_afip import models


def get_test_file(filename: str, mode="r") -> Path:
    """Helper to get test files."""
    path = Path(__file__).parent / "testing" / filename
    return path


class UserFactory(DjangoModelFactory):
    class Meta:
        model = User

    username = "john doe"
    email = "john@doe.co"
    password = PostGenerationMethodCall("set_password", "123")


class SuperUserFactory(UserFactory):
    is_staff = True
    is_superuser = True


class ConceptTypeFactory(DjangoModelFactory):
    class Meta:
        model = models.ConceptType
        django_get_or_create = ("code",)

    code = "1"
    description = "Producto"
    valid_from = date(2010, 9, 17)


class DocumentTypeFactory(DjangoModelFactory):
    class Meta:
        model = models.DocumentType

    code = "96"
    description = "DNI"
    valid_from = date(2008, 7, 25)


class CurrencyTypeFactory(DjangoModelFactory):
    class Meta:
        model = models.CurrencyType

    code = "PES"
    description = "Pesos Argentinos"
    valid_from = date(2009, 4, 3)


class ReceiptTypeFactory(DjangoModelFactory):
    class Meta:
        model = models.ReceiptType
        django_get_or_create = ("code",)

    code = "11"
    description = "Factura C"
    valid_from = date(2011, 3, 30)


class TaxPayerFactory(DjangoModelFactory):
    class Meta:
        model = models.TaxPayer
        django_get_or_create = ["cuit"]

    name = "John Smith"
    cuit = 20329642330
    is_sandboxed = True
    key = FileField(from_path=get_test_file("test.key"))
    certificate = FileField(from_path=get_test_file("test.crt"))
    active_since = datetime(2011, 10, 3)
    logo = ImageField(from_path=get_test_file("tiny.png", "rb"))


class AlternateTaxpayerFactory(DjangoModelFactory):
    """A taxpayer with an alternate (valid) keypair."""

    class Meta:
        model = models.TaxPayer

    name = "John Smith"
    cuit = 20329642330
    is_sandboxed = True
    key = FileField(from_path=get_test_file("test2.key"))
    certificate = FileField(from_path=get_test_file("test2.crt"))
    active_since = datetime(2011, 10, 3)


class PointOfSalesFactory(DjangoModelFactory):
    class Meta:
        model = models.PointOfSales
        django_get_or_create = ["owner", "number"]

    number = 1
    issuance_type = "CAE"
    blocked = False
    owner = SubFactory(TaxPayerFactory)
    # TODO: Renamethis to something more regional:
    issuing_name = "Red Company Inc."
    issuing_address = "100 Red Av\nRedsville\nUK"
    issuing_email = "billing@example.com"
    vat_condition = "Exempt"
    gross_income_condition = "Exempt"
    sales_terms = "Credit Card"


class ReceiptFactory(DjangoModelFactory):
    class Meta:
        model = models.Receipt

    concept = SubFactory(ConceptTypeFactory, code=1)
    document_type = SubFactory(DocumentTypeFactory, code=96)
    document_number = 203012345
    issued_date = LazyFunction(date.today)
    total_amount = 130
    net_untaxed = 0
    net_taxed = 100
    exempt_amount = 0
    currency = SubFactory(CurrencyTypeFactory, code="PES")
    currency_quote = 1
    receipt_type = SubFactory(ReceiptTypeFactory, code=6)
    point_of_sales = SubFactory(PointOfSalesFactory)


class ReceiptWithVatAndTaxFactory(ReceiptFactory):
    """Receipt with a valid Vat and Tax, ready to validate."""

    point_of_sales = LazyFunction(lambda: models.PointOfSales.objects.first())

    @post_generation
    def post(obj: models.Receipt, create, extracted, **kwargs):
        VatFactory(vat_type__code=5, receipt=obj)
        TaxFactory(tax_type__code=3, receipt=obj)


class ReceiptWithInconsistentVatAndTaxFactory(ReceiptWithVatAndTaxFactory):
    """Receipt with a valid Vat and Tax, ready to validate."""

    document_type = SubFactory(DocumentTypeFactory, code=80)

    @post_generation
    def post(obj: models.Receipt, create, extracted, **kwargs):
        VatFactory(vat_type__code=5, receipt=obj)
        TaxFactory(tax_type__code=3, receipt=obj)


class ReceiptWithApprovedValidation(ReceiptFactory):
    """Receipt with fake (e.g.: not live) approved validation."""

    receipt_number = Sequence(lambda n: n + 1)

    @post_generation
    def post(obj: models.Receipt, create, extracted, **kwargs):
        ReceiptValidationFactory(receipt=obj)


class ReceiptValidationFactory(DjangoModelFactory):
    class Meta:
        model = models.ReceiptValidation

    result = models.ReceiptValidation.RESULT_APPROVED
    processed_date = make_aware(datetime(2017, 7, 2, 21, 6, 4))
    cae = "67190616790549"
    cae_expiration = make_aware(datetime(2017, 7, 12))
    receipt = SubFactory(ReceiptFactory, receipt_number=17)


class ReceiptPDFFactory(DjangoModelFactory):
    class Meta:
        model = models.ReceiptPDF

    client_address = "La Rioja 123\nX5000EVX Córdoba"
    client_name = "John Doe"
    client_vat_condition = "Consumidor Final"
    gross_income_condition = "Convenio Multilateral"
    issuing_address = "Happy Street 123, CABA"
    issuing_name = "Alice Doe"
    receipt = SubFactory(ReceiptFactory)
    sales_terms = "Contado"
    vat_condition = "Responsable Monotributo"


class ReceiptPDFWithFileFactory(ReceiptPDFFactory):
    receipt = SubFactory(ReceiptWithApprovedValidation)


class GenericAfipTypeFactory(DjangoModelFactory):
    class Meta:
        model = models.GenericAfipType

    code = 80
    description = "CUIT"
    valid_from = datetime(2017, 8, 10)


class VatTypeFactory(GenericAfipTypeFactory):
    class Meta:
        model = models.VatType

    code = 5
    description = "21%"


class TaxTypeFactory(GenericAfipTypeFactory):
    class Meta:
        model = models.TaxType


class VatFactory(DjangoModelFactory):
    class Meta:
        model = models.Vat

    amount = 21
    base_amount = 100
    receipt = SubFactory(ReceiptFactory)
    vat_type = SubFactory(VatTypeFactory)


class TaxFactory(DjangoModelFactory):
    class Meta:
        model = models.Tax

    aliquot = 9
    amount = 9
    base_amount = 100
    receipt = SubFactory(ReceiptFactory)
    tax_type = SubFactory(TaxTypeFactory)



# Mypy doesn't play well with action-functions.
#
# type: ignore
import logging
from contextlib import contextmanager
from datetime import datetime

from django.contrib import admin
from django.contrib import messages
from django.db.models import F
from django.http import HttpResponse
from django.urls import reverse
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _

from django_afip import exceptions
from django_afip import models

logger = logging.getLogger(__name__)


# TODO: Add an action to populate generic types.


@contextmanager
def catch_errors(self, request):
    """Catches specific errors in admin actions and shows a friendly error."""
    try:
        yield
    except exceptions.CertificateExpired:
        self.message_user(
            request,
            _("The AFIP Taxpayer certificate has expired."),
            messages.ERROR,
        )
    except exceptions.UntrustedCertificate:
        self.message_user(
            request,
            _("The AFIP Taxpayer certificate is untrusted."),
            messages.ERROR,
        )
    except exceptions.CorruptCertificate:
        self.message_user(
            request,
            _("The AFIP Taxpayer certificate is corrupt."),
            messages.ERROR,
        )
    except exceptions.AuthenticationError as e:
        logger.exception("AFIP auth failed")
        self.message_user(
            request,
            _("An unknown authentication error has ocurred: %s") % e,
            messages.ERROR,
        )


class VatInline(admin.TabularInline):
    model = models.Vat
    fields = (
        "vat_type",
        "base_amount",
        "amount",
    )
    extra = 1


class TaxInline(admin.TabularInline):
    model = models.Tax
    fields = (
        "tax_type",
        "description",
        "base_amount",
        "aliquot",
        "amount",
    )
    extra = 1


class ReceiptEntryInline(admin.TabularInline):
    model = models.ReceiptEntry
    fields = (
        "description",
        "quantity",
        "unit_price",
        "vat",
    )
    extra = 1


class ReceiptValidationInline(admin.StackedInline):
    model = models.ReceiptValidation
    readonly_fields = (
        "result",
        "processed_date",
        "cae",
        "cae_expiration",
        "observations",
    )
    extra = 0


class ReceiptStatusFilter(admin.SimpleListFilter):
    title = _("status")
    parameter_name = "status"

    VALIDATED = "validated"
    NOT_VALIDATED = "not_validated"

    def lookups(self, request, model_admin):
        return (
            (self.VALIDATED, _("Validated")),
            (self.NOT_VALIDATED, _("Not validated")),
        )

    def queryset(self, request, queryset):
        if self.value() == self.VALIDATED:
            return queryset.filter(
                validation__result=models.ReceiptValidation.RESULT_APPROVED
            )
        if self.value() == self.NOT_VALIDATED:
            return queryset.exclude(
                validation__result=models.ReceiptValidation.RESULT_APPROVED
            )


class ReceiptTypeFilter(admin.SimpleListFilter):
    title = models.ReceiptType._meta.verbose_name
    parameter_name = "receipt_type"

    def lookups(self, request, model_admin):
        return (
            (receipt_type.code, receipt_type.description)
            for receipt_type in models.ReceiptType.objects.filter(
                receipts__isnull=False,
            ).distinct()
        )

    def queryset(self, request, queryset):
        value = self.value()
        if value:
            queryset = queryset.filter(receipt_type__code=value)
        return queryset


@admin.register(models.Receipt)
class ReceiptAdmin(admin.ModelAdmin):
    search_fields = ("receipt_number",)
    list_display = (
        "id",
        "receipt_type",
        "point_of_sales",
        "number",
        "issued_date",
        "friendly_total_amount",
        "validated",
        "pdf_link",
    )
    list_filter = (
        ReceiptStatusFilter,
        ReceiptTypeFilter,
    )
    autocomplete_fields = (
        "currency",
        "receipt_type",
        "related_receipts",
    )
    date_hierarchy = "issued_date"

    __related_fields = [
        "validated",
        "cae",
    ]

    inlines = (
        VatInline,
        TaxInline,
        ReceiptEntryInline,
        ReceiptValidationInline,
    )
    ordering = ("-issued_date",)

    readonly_fields = __related_fields

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .select_related(
                "receipt_type",
                "point_of_sales",
            )
            .annotate(
                pdf_id=F("receiptpdf__id"),
                validation_result=F("validation__result"),
            )
        )

    def number(self, obj):
        return obj.formatted_number

    number.short_description = _("receipt number")
    number.admin_order_field = "receipt_number"

    def friendly_total_amount(self, obj):
        return "{:0.2f} ARS{}".format(
            obj.total_amount * obj.currency_quote,
            "*" if obj.currency_quote != 1 else "",
        )

    friendly_total_amount.short_description = _("total amount")

    def validated(self, obj):
        return obj.validation_result == models.ReceiptValidation.RESULT_APPROVED

    validated.short_description = _("validated")
    validated.admin_order_field = "validation__result"
    validated.boolean = True

    def pdf_link(self, obj):
        if not obj.pdf_id:
            return mark_safe(
                '<a href="{}?receipt={}">{}</a>'.format(
                    reverse(self.admin_site.name + ":afip_receiptpdf_add"),
                    obj.id,
                    _("Create"),
                )
            )
        return mark_safe(
            '<a href="{}">{}</a>'.format(
                reverse(
                    self.admin_site.name + ":afip_receiptpdf_change",
                    args=(obj.pdf_id,),
                ),
                _("Edit"),
            )
        )

    pdf_link.short_description = _("PDF")
    pdf_link.admin_order_field = "receiptpdf__id"

    def cae(self, obj):
        return obj.validation.cae

    cae.short_description = _("cae")
    cae.admin_order_field = "validation__cae"

    def validate(self, request, queryset):
        with catch_errors(self, request):
            errs = queryset.validate()

        if errs:
            self.message_user(
                request,
                _("Receipt validation failed: %s.") % errs,
                messages.ERROR,
            )

    validate.short_description = _("Validate")

    actions = [validate]


@admin.register(models.AuthTicket)
class AuthTicketAdmin(admin.ModelAdmin):
    list_display = (
        "unique_id",
        "owner",
        "service",
        "generated",
        "expires",
    )


@admin.register(models.TaxPayer)
class TaxPayerAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "cuit",
        "certificate_expiration",
    )

    def fetch_points_of_sales(self, request, queryset):
        with catch_errors(self, request):
            poses = [
                pos
                for taxpayer in queryset.all()
                for pos in taxpayer.fetch_points_of_sales()
            ]

        created = len([pos for pos in poses if pos[1]])
        skipped = len(poses) - created

        self.message_user(
            request,
            message=(_("%d points of sales created.") % created),
            level=messages.SUCCESS,
        )
        self.message_user(
            request,
            message=(_("%d points of sales already existed.") % skipped),
            level=messages.WARNING,
        )

    fetch_points_of_sales.short_description = _("Fetch points of sales")

    def generate_key(self, request, queryset):
        key_count = sum(t.generate_key() for t in queryset.all())

        if key_count == 1:
            message = _("Key generated successfully.")
            level = messages.SUCCESS
        elif key_count:
            message = _("%d keys generated successfully.") % key_count
            level = messages.SUCCESS
        else:
            message = _("No keys generated; Taxpayers already had keys.")
            level = messages.ERROR

        self.message_user(
            request,
            message=message,
            level=level,
        )

    generate_key.short_description = _("Generate key")

    def generate_csr(self, request, queryset):
        if queryset.count() > 1:
            self.message_user(
                request,
                message=_("Can only generate CSR for one taxpayer at a time."),
                level=messages.ERROR,
            )
            return

        taxpayer = queryset.first()
        if not taxpayer.key:
            taxpayer.generate_key()

        csr = taxpayer.generate_csr()
        filename = "cuit-{}-{}.csr".format(
            taxpayer.cuit,
            int(datetime.now().timestamp()),
        )

        response = HttpResponse(content_type="application/pkcs10")
        response["Content-Disposition"] = f"attachment; filename={filename}"

        response.write(csr.read())
        return response

    generate_csr.short_description = _("Generate CSR")

    actions = (
        fetch_points_of_sales,
        generate_key,
        generate_csr,
    )


@admin.register(models.PointOfSales)
class PointOfSalesAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "owner",
        "number",
        "issuance_type",
        "blocked",
        "drop_date",
    )


@admin.register(models.CurrencyType)
class CurrencyTypeAdmin(admin.ModelAdmin):
    search_fields = (
        "code",
        "description",
    )
    list_display = (
        "code",
        "description",
        "valid_from",
        "valid_to",
    )


@admin.register(models.ReceiptType)
class ReceiptTypeAdmin(admin.ModelAdmin):
    search_fields = (
        "code",
        "description",
    )
    list_display = (
        "code",
        "description",
        "valid_from",
        "valid_to",
    )


class ReceiptHasFileFilter(admin.SimpleListFilter):
    title = _("has file")
    parameter_name = "has_file"

    YES = "yes"
    NO = "no"

    def lookups(self, request, model_admin):
        return (
            (self.YES, _("Yes")),
            (self.NO, _("No")),
        )

    def queryset(self, request, queryset):
        if self.value() == self.YES:
            return queryset.exclude(pdf_file="")
        if self.value() == self.NO:
            return queryset.filter(pdf_file="")
        return queryset


@admin.register(models.ReceiptPDF)
class ReceiptPDFAdmin(admin.ModelAdmin):
    list_display = (
        "receipt_id",
        "taxpayer",
        "receipt",
        "client_name",
        "has_file",
    )
    list_filter = (
        ReceiptHasFileFilter,
        "receipt__point_of_sales__owner",
        "receipt__receipt_type",
    )
    raw_id_fields = ("receipt",)

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .select_related(
                "receipt",
                "receipt__point_of_sales__owner",
                "receipt__receipt_type",
            )
        )

    def taxpayer(self, obj):
        return obj.receipt.point_of_sales.owner

    taxpayer.short_description = models.TaxPayer._meta.verbose_name

    def has_file(self, obj):
        return bool(obj.pdf_file)

    has_file.admin_order_field = "pdf_file"
    has_file.boolean = True
    has_file.short_description = _("Has file")

    def generate_pdf(self, request, queryset):
        for pdf in queryset:
            pdf.save_pdf()

    generate_pdf.short_description = _("Generate pdf")

    actions = (generate_pdf,)


@admin.register(models.ReceiptValidation)
class ReceiptValidationAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "receipt_number",
        "successful",
        "cae",
        "processed_date",
    )

    raw_id_fields = ("receipt",)

    def receipt_number(self, obj):
        return obj.receipt.formatted_number

    receipt_number.short_description = _("receipt number")
    receipt_number.admin_order_field = "receipt_id"

    def successful(self, obj):
        return obj.result == models.ReceiptValidation.RESULT_APPROVED

    successful.short_description = _("result")
    successful.admin_order_field = "result"
    successful.boolean = True


admin.site.register(models.ConceptType)
admin.site.register(models.DocumentType)
admin.site.register(models.VatType)
admin.site.register(models.TaxType)
admin.site.register(models.Observation)



from django.utils.functional import cached_property
from django_renderpdf.views import PDFView

from django_afip import models
from django_afip.pdf import get_encoded_qrcode

# Note: When updating this, be sure to update the docstring of the method that uses
# these below.
TEMPLATE_NAMES = [
    "receipts/{taxpayer}/pos_{point_of_sales}/code_{code}.html",
    "receipts/{taxpayer}/code_{code}.html",
    "receipts/code_{code}.html",
    "receipts/{code}.html",
]


class ReceiptPDFView(PDFView):
    @cached_property
    def receipt(self):
        return models.Receipt.objects.select_related(
            "receipt_type",
            "point_of_sales",
        ).get(
            pk=self.kwargs["pk"],
        )

    @property
    def download_name(self):
        return f"{self.receipt.formatted_number}.pdf"

    def get_template_names(self, receipt: models.Receipt = None):
        """Return the templates use to render the Receipt PDF.

        Template discovery tries to find any of the below receipts::

            receipts/{taxpayer}/pos_{point_of_sales}/code_{code}.html
            receipts/{taxpayer}/code_{code}.html
            receipts/code_{code}.html
            receipts/{code}.html

        For example, to override the "Factura C" template for point of sales 0002 for
        Taxpayer 20-32964233-0, use::

            receipts/20329642330/pos_2/code_6.html
        """
        receipt = receipt or self.receipt
        assert receipt is not None

        return [
            template.format(
                taxpayer=receipt.point_of_sales.owner.cuit,
                point_of_sales=receipt.point_of_sales.number,
                code=receipt.receipt_type.code,
            )
            for template in TEMPLATE_NAMES
        ]

    @staticmethod
    def get_context_for_pk(pk, *args, **kwargs):
        context = {}

        receipt_pdf = (
            models.ReceiptPDF.objects.select_related(
                "receipt",
                "receipt__receipt_type",
                "receipt__document_type",
                "receipt__validation",
                "receipt__point_of_sales",
                "receipt__point_of_sales__owner",
            )
            .prefetch_related(
                "receipt__entries",
            )
            .get(
                receipt__pk=pk,
            )
        )

        # Prefetch required data in a single query:
        receipt_pdf.receipt = (
            models.Receipt.objects.select_related(
                "receipt_type",
                "document_type",
                "validation",
                "point_of_sales",
                "point_of_sales__owner",
            )
            .prefetch_related(
                "entries",
            )
            .get(
                pk=receipt_pdf.receipt_id,
            )
        )
        taxpayer = receipt_pdf.receipt.point_of_sales.owner

        context["pdf"] = receipt_pdf
        context["taxpayer"] = taxpayer
        context["qrcode"] = get_encoded_qrcode(receipt_pdf)

        return context

    def get_context_data(self, *args, pk=None, **kwargs):
        context = super().get_context_data(*args, pk=pk, **kwargs)
        context.update(self.get_context_for_pk(pk, *args, **kwargs))
        return context



#!/usr/bin/env python3
from setuptools import find_packages
from setuptools import setup

setup(
    name="django-afip",
    description="AFIP integration for django",
    author="Hugo Osvaldo Barrera",
    author_email="hugo@barrera.io",
    url="https://github.com/WhyNotHugo/django-afip",
    project_urls={
        "Documentation": "https://django-afip.readthedocs.io/",
        "Issue Tracker": "https://github.com/WhyNotHugo/django-afip/issues",
        "Donate": "https://liberapay.com/WhyNotHugo/",
    },
    license="ISC",
    packages=find_packages(),
    include_package_data=True,
    long_description=open("README.rst").read(),
    install_requires=[
        "cryptography>=3.2,<4.0",
        "django>=3.1,<3.3",
        "django_renderpdf>=3.0.0,<4.0.0",
        "lxml>=3.4.4",
        "pyopenssl>=16.2.0",
        "pytz>=2015.4",
        "setuptools-git>=1.1",
        "setuptools-scm>=1.7.0",
        "wheel>=0.24.0",
        "zeep>=1.1.0,<5.0.0",
        "qrcode[pil]>=6.1,<7.0",
        "pyyaml>=5.3.1,<6.0.0",
    ],
    extras_require={
        "docs": ["Sphinx", "sphinx-autobuild"],
        "postgres": ["psycopg2"],
        "mysql": ["mysqlclient"],
        "factories": ["factory-boy"],
    },
    use_scm_version={
        "version_scheme": "post-release",
        "write_to": "django_afip/version.py",
    },
    setup_requires=["setuptools_scm"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 3.0",
        "Framework :: Django :: 3.1",
        "Framework :: Django :: 3.2",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: ISC License (ISCL)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



# Generated by Django 2.1.5 on 2019-01-29 04:36
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ("afip", "0002_taxpayerextras"),
    ]

    operations = [
        migrations.AlterField(
            model_name="pointofsales",
            name="issuance_type",
            field=models.CharField(
                help_text="Indicates if this POS emits using CAE and CAEA.",
                max_length=24,
                verbose_name="issuance type",
            ),
        ),
    ]



__all__ = ("get_client",)

from urllib.parse import urlparse

import pytz
from django.utils.functional import LazyObject
from requests import Session
from requests.adapters import HTTPAdapter
from urllib3.util.ssl_ import DEFAULT_CIPHERS
from urllib3.util.ssl_ import create_urllib3_context
from zeep import Client
from zeep.cache import SqliteCache
from zeep.transports import Transport

TZ_AR = pytz.timezone(pytz.country_timezones["ar"][0])
CIPHERS = DEFAULT_CIPHERS + "HIGH:!DH:!aNULL"
WSDLS = {
    ("wsaa", False): "https://wsaa.afip.gov.ar/ws/services/LoginCms?wsdl",
    ("wsfe", False): "https://servicios1.afip.gov.ar/wsfev1/service.asmx?WSDL",
    ("wsaa", True): "https://wsaahomo.afip.gov.ar/ws/services/LoginCms?wsdl",
    ("wsfe", True): "https://wswhomo.afip.gov.ar/wsfev1/service.asmx?WSDL",
}


class AFIPAdapter(HTTPAdapter):
    """An adapter with reduced security so it'll work with AFIP."""

    def init_poolmanager(self, *args, **kwargs):
        context = create_urllib3_context(ciphers=CIPHERS)
        kwargs["ssl_context"] = context
        return super().init_poolmanager(*args, **kwargs)

    def proxy_manager_for(self, *args, **kwargs):
        context = create_urllib3_context(ciphers=CIPHERS)
        kwargs["ssl_context"] = context
        return super().proxy_manager_for(*args, **kwargs)


class LazyTransport(LazyObject):
    """A lazy-initialized Zeep transport.

    This transport does two non-default things:
    - Reduces TLS security. Sadly, AFIP only has insecure endpoints, so we're forced to
      reduce security to talk to them.
    - Cache the WSDL file for a whole day.
    """

    def _setup(self):
        """Initialise this lazy object with a celery app instance."""
        session = Session()

        # For each WSDL, extract the domain, and add it as an exception:
        for url in WSDLS.values():
            parsed = urlparse(url)
            base_url = f"{parsed.scheme}://{parsed.netloc}"
            session.mount(base_url, AFIPAdapter())

        self._wrapped = Transport(cache=SqliteCache(timeout=86400), session=session)


transport = LazyTransport()
cached_clients = {}


def get_client(service_name: str, sandbox=False) -> Client:
    """
    Return a client for a given service.

    The `sandbox` argument should only be necessary if a the client will be
    used to make a request. If it will only be used to serialize objects, it is
    irrelevant. A caller can avoid the overhead of determining the sandbox mode in the
    calling context if only serialization operations will take place.

    :param service_name: The name of the web services.
    :param sandbox: Whether the sandbox (or production) environment should
        be used by the returned client.
    :returns: A zeep client to communicate with an AFIP web service.
    """
    key = (
        service_name.lower(),
        sandbox,
    )

    try:
        if key not in cached_clients:
            cached_clients[key] = Client(WSDLS[key], transport=transport)

        return cached_clients[key]
    except KeyError:
        raise ValueError(f"Unknown service name, {service_name}")



from django.apps import AppConfig


class AfipConfig(AppConfig):
    name = "django_afip"
    label = "afip"
    verbose_name = "AFIP"
    default_auto_field = "django.db.models.AutoField"

    def ready(self):
        # Register app signals:
        from django_afip import signals  # noqa: F401



import pytest
from django.core import management

from django_afip.models import GenericAfipType


@pytest.mark.django_db
def test_afip_metadata_command():
    assert len(GenericAfipType.SUBCLASSES) == 6

    for model in GenericAfipType.SUBCLASSES:
        assert model.objects.count() == 0

    management.call_command("afipmetadata")

    for model in GenericAfipType.SUBCLASSES:
        assert model.objects.count() > 0



from django.utils.functional import LazyObject

from django_afip.clients import get_client


class _LazyFactory(LazyObject):
    """A lazy-initialised factory for WSDL objects."""

    def _setup(self):
        self._wrapped = get_client("wsfe").type_factory("ns0")


f = _LazyFactory()


def serialize_datetime(datetime):
    """
    "Another date formatting function?" you're thinking, eh? Well, this
    actually formats dates in the *exact* format the AFIP's WS expects it,
    which is almost like ISO8601.

    Note that .isoformat() works fine on production servers, but not on the
    sandbox ones.
    """
    return datetime.strftime("%Y-%m-%dT%H:%M:%S-00:00")


def serialize_date(date):
    return date.strftime("%Y%m%d")


def serialize_ticket(ticket):
    return f.FEAuthRequest(
        Token=ticket.token,
        Sign=ticket.signature,
        Cuit=ticket.owner.cuit,
    )


def serialize_multiple_receipts(receipts):
    receipts = receipts.all().order_by("receipt_number")

    first = receipts.first()
    receipts = [serialize_receipt(receipt) for receipt in receipts]

    serialised = f.FECAERequest(
        FeCabReq=f.FECAECabRequest(
            CantReg=len(receipts),
            PtoVta=first.point_of_sales.number,
            CbteTipo=first.receipt_type.code,
        ),
        FeDetReq=f.ArrayOfFECAEDetRequest(receipts),
    )

    return serialised


def serialize_receipt(receipt):
    taxes = receipt.taxes.all()
    vats = receipt.vat.all()

    serialized = f.FECAEDetRequest(
        Concepto=receipt.concept.code,
        DocTipo=receipt.document_type.code,
        DocNro=receipt.document_number,
        # TODO: Check that this is not None!,
        CbteDesde=receipt.receipt_number,
        CbteHasta=receipt.receipt_number,
        CbteFch=serialize_date(receipt.issued_date),
        ImpTotal=receipt.total_amount,
        ImpTotConc=receipt.net_untaxed,
        ImpNeto=receipt.net_taxed,
        ImpOpEx=receipt.exempt_amount,
        ImpIVA=sum(vat.amount for vat in vats),
        ImpTrib=sum(tax.amount for tax in taxes),
        MonId=receipt.currency.code,
        MonCotiz=receipt.currency_quote,
    )
    if int(receipt.concept.code) in (2, 3):
        serialized.FchServDesde = serialize_date(receipt.service_start)
        serialized.FchServHasta = serialize_date(receipt.service_end)
        serialized.FchVtoPago = serialize_date(receipt.expiration_date)

    if taxes:
        serialized.Tributos = f.ArrayOfTributo([serialize_tax(tax) for tax in taxes])

    if vats:
        serialized.Iva = f.ArrayOfAlicIva([serialize_vat(vat) for vat in vats])

    related_receipts = receipt.related_receipts.all()
    if related_receipts:
        serialized.CbtesAsoc = f.ArrayOfCbteAsoc(
            [
                f.CbteAsoc(
                    r.receipt_type.code,
                    r.point_of_sales.number,
                    r.receipt_number,
                )
                for r in related_receipts
            ]
        )

    return serialized


def serialize_tax(tax):
    return f.Tributo(
        Id=tax.tax_type.code,
        Desc=tax.description,
        BaseImp=tax.base_amount,
        Alic=tax.aliquot,
        Importe=tax.amount,
    )


def serialize_vat(vat):
    return f.AlicIva(
        Id=vat.vat_type.code,
        BaseImp=vat.base_amount,
        Importe=vat.amount,
    )


def serialize_receipt_data(receipt_type, receipt_number, point_of_sales):
    return f.FECompConsultaReq(
        CbteTipo=receipt_type, CbteNro=receipt_number, PtoVta=point_of_sales
    )
